#include <GL/glut.h>
#include "octree.h"
#include "octreeLoader.h"
#include "frustum.h"
#include <cstdio>
#include <vector>
#include <set>
#include <deque>
#include <stack>

using namespace std;


void getLastError();
//bool OctreeLoader::conservativeMode=false;

unsigned int OctreeLoader::loadedVertexCounter=0;
unsigned int OctreeLoader::loadedFaceCounter=0;
OctreeNodeSet OctreeLoader::loadedNodeSet;
//HANDLE OctreeLoader::loadingThread=NULL;
//HANDLE OctreeLoader::mutex=NULL;
//HANDLE OctreeLoader::event=NULL;
HANDLE OctreeLoader::vertexMapHandle=NULL;
HANDLE OctreeLoader::faceMapHandle=NULL;
//OctreeNodeQueue OctreeLoader::loadingQueue;

//CRITICAL_SECTION OctreeLoader::criticalSection;

//bool OctreeLoader::canTerminate=false;

int OctreeLoader::currentIndex=0;
int OctreeLoader::nextIndex=1;
Octree* OctreeLoader::root=NULL;
OctreeNodeSet OctreeLoader::nodeSet[MaxCoherentFrameNumber];

//int OctreeLoader::ignoredLevel=12;

int OctreeLoader::capacityFactor=4096;


/*
DWORD WINAPI loadingProc(LPVOID param)
{	
	while (!OctreeLoader::canTerminate)
	{		
		if (WaitForSingleObject(OctreeLoader::mutex, INFINITE)!=WAIT_OBJECT_0)
		{
			printf("wait error %d\n", GetLastError());
			exit(8);
		}
		if (OctreeLoader::loadingQueue.empty())
		{		
			if (!ReleaseMutex(OctreeLoader::mutex))
			{
				printf("release error %d\n", GetLastError());
				exit(8);
			}
			continue;
		}			
		OctreeLoader::mapOctreeNode(OctreeLoader::loadingQueue.front());
		OctreeLoader::loadingQueue.pop_front();
		if (!ReleaseMutex(OctreeLoader::mutex))
		{
			printf("release error %d\n", GetLastError());
			exit(8);
		}	
		
	}
	return 0;
}
*/





void vectorMultiply(VVector v1, VVector v2, VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=v1[i]*v2[i];
	}
}


void vectorScalor(VVector v1, VType scalor)
{
	for (int i=0; i<3; i++)
	{
		v1[i]*=scalor;
	}
}

void vectorScalor(VVector v1, VType scalor, VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=v1[i]*scalor;
	}
}

void vectorMinus(VVector v1, VVector v2, VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=v1[i]-v2[i];
	}
}

void vectorPlus(VVector v1, VVector v2, VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=v1[i]+v2[i];
	}
}

void vectorAssign(VVector source, VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=source[i];
	}
}

//v1=v0 + ops*offset
//v0=v1-ops*offset
void reverseCalcCenter(VVector subCenter, VType width, VVector operations, VVector result)
{
	vectorScalor(operations, (-1.0)*width, result);
	vectorPlus(subCenter, result, result);
}

void calcRelativeOperation(VVector refOps, VVector oldOps, VVector relativeOps)
{
	vectorMultiply(refOps, oldOps, relativeOps);
}

//calc image coord of 'coord' with respect to center
//v1 is coord given, v2 is what we want to calculate,
//v0 is center and offset is v1-v0
//let me explain as following:
//  v1= v0 + op1 * offset
//	v2= v0 + op2 * offset
//  op1* op1= op2*op2 = IdentityVector
//  v2=v0+ op2*( op1*op1*offset)=v0+op2*(op1*(v1-v0))= v0 + (v1-v0)*op2*op1
//so relative operation =op1*op2
void calcCenterImage(VVector center, VVector coord, VVector relativeOps, VVector result)
{
	VVector delta;
	vectorMinus(coord, center, delta);
	vectorMultiply(delta, relativeOps, result);
	vectorPlus(center, result, result);
}

void calcShiftImage(VVector coord, VVector relativeOps, float offset,  VVector result)
{
	for (int i=0; i<3; i++)
	{
		result[i]=coord[i]+relativeOps[i]*offset;
	}
}


void calcShiftedLeafNode(Octree* refOctree, OctreeShiftDescriptor& octreeShift, Octree* result)
{
	int i;
	memcpy(result, refOctree, sizeof(Octree));
	result->level=refOctree->level+octreeShift.levelOffset;

	calcCenterImage(octreeShift.newCenter, refOctree->center, octreeShift.relativeOps, result->center);

	result->myNodeIndex=refOctree->myNodeIndex+octreeShift.indexOffset;
	result->parentIndex=refOctree->parentIndex+octreeShift.indexOffset;

	result->vStartPageIndex=refOctree->vStartPageIndex+octreeShift.vPageOffset;
	result->fStartPageIndex=refOctree->fStartPageIndex+octreeShift.fPageOffset;

	//surely you have noticed that fileAddLow and fileAddHigh are redundent data since
	//we already have the startpageIndex, but the whole point is to use space for speed.
	result->calculateFileOffsetV(result->vStartPageIndex, 0, result->vFileAddLow, result->vFileAddHigh);
	result->calculateFileOffsetF(result->fStartPageIndex, 0, result->fFileAddLow, result->fFileAddHigh);

	for (i=0; i<8; i++)
	{
		if (refOctree->sons[i]!=-1)
		{
			result->sons[i]=refOctree->sons[i]+octreeShift.indexOffset;
		}	
	}

	for (i=0; i<refOctree->neighbourCount; i++)
	{
		result->neighbours[i]=refOctree->neighbours[i]+octreeShift.indexOffset;
	}
}

/*
//raw data just needs to be copied after it is transformed
void OctreeLoader::copyOctreeLeafNode(Octree* refOctree, Octree* targetOctree, OctreeShiftDescriptor& octreeShift)
{
	void* source;
	int i;
        PlyVertex* refVertex;
        PlyVertex  targetVertex;
        //FaceStruct* refFace;
        //FaceStruct targetFace;
        source=GenericMapViewOfFile(vertexMapHandle, true, refOctree->vFileAddHigh, refOctree->vFileAddLow,
                refOctree->vMapSize);                

        GenericSetFilePointer(Octree::vertexHandle, targetOctree->vFileAddLow, targetOctree->vFileAddHigh);
        refVertex=(PlyVertex*) source;
        for (i=0; i<refOctree->vNumber+refOctree->vRepeatNumber; i++)
        {
		printf("before memcpy\n");
                memcpy(targetVertex.color, refVertex[i].color, sizeof(UCVector));
                memcpy(targetVertex.normal, refVertex[i].normal, sizeof(VVector));
		
                //calcCenterImage(octreeShift.newCenter, refVertex[i].coord, octreeShift.relativeOps, 
		//	targetVertex.coord);
		
		calcShiftImage(refVertex[i].coord, octreeShift.relativeOps, octreeShift.vertexOffset, 
			targetVertex.coord);

                GenericWriteFile(Octree::vertexHandle, &targetVertex, sizeof(PlyVertex));
		printf("after one copy\n");
        }
        GenericUnmapViewOfFile(source, refOctree->vMapSize);



        source=GenericMapViewOfFile(faceMapHandle, true, refOctree->fFileAddHigh, refOctree->fFileAddLow,
                refOctree->fMapSize);
        GenericSetFilePointer(Octree::faceHandle, targetOctree->fFileAddLow, targetOctree->fFileAddHigh);
        GenericWriteFile(Octree::faceHandle, source, refOctree->fMapSize);

        GenericUnmapViewOfFile(source, refOctree->fMapSize);
}
*/


//raw data just needs to be copied after it is transformed
void OctreeLoader::copyOctreeLeafNode(Octree* refOctree, Octree* targetOctree, OctreeShiftDescriptor& octreeShift)
{
	void* source;
	void* target;
	int i;
	PlyVertex* refVertex;
	PlyVertex* targetVertex;
	//FaceStruct* refFace;
	//FaceStruct targetFace;
	//printf("before mapping to both nodes\n");
	//refOctree->display();
	//targetOctree->display();


	source=GenericMapViewOfFile(vertexMapHandle, false, refOctree->vFileAddHigh, refOctree->vFileAddLow,  
		refOctree->vMapSize);
	target=GenericMapViewOfFile(vertexMapHandle, false, targetOctree->vFileAddHigh, targetOctree->vFileAddLow, 
		targetOctree->vMapSize);
	//printf("after file mapping both nodes\n");
	//GenericSetFilePointer(Octree::vertexHandle, targetOctree->vFileAddLow, targetOctree->vFileAddHigh);
	refVertex=(PlyVertex*) source;
	targetVertex=(PlyVertex*)target;
	for (i=0; i<refOctree->vNumber+refOctree->vRepeatNumber; i++)
	{
		memcpy(targetVertex[i].color, refVertex[i].color, sizeof(UCVector));
		
		memcpy(targetVertex[i].normal, refVertex[i].normal, sizeof(VVector));
		//printf("after memcopy\n");
		//calcCenterImage(octreeShift.newCenter, refVertex[i].coord, octreeShift.relativeOps, 
		//	targetVertex[i].coord);
		//GenericWriteFile(Octree::vertexHandle, &targetVertex, sizeof(PlyVertex));
		calcShiftImage(refVertex[i].coord, octreeShift.relativeOps, octreeShift.vertexOffset, 
			targetVertex[i].coord);
	}
	//printf("after copying data\n");
	GenericUnmapViewOfFile(source, refOctree->vMapSize);
	GenericUnmapViewOfFile(target, targetOctree->vMapSize);



	source=GenericMapViewOfFile(faceMapHandle, true, refOctree->fFileAddHigh, refOctree->fFileAddLow, 
		refOctree->fMapSize);
	target=GenericMapViewOfFile(faceMapHandle, false, targetOctree->fFileAddHigh, targetOctree->fFileAddLow,
		targetOctree->fMapSize);

	//GenericSetFilePointer(Octree::faceHandle, targetOctree->fFileAddLow, targetOctree->fFileAddHigh);
	//GenericWriteFile(Octree::faceHandle, source, refOctree->fMapSize);
	memcpy(target, source, refOctree->fMapSize);
	
	GenericUnmapViewOfFile(source, refOctree->fMapSize);
	GenericUnmapViewOfFile(target, targetOctree->fMapSize);
}

void OctreeLoader::calcOctreeShiftDescriptor(OctreeShiftDescriptor& octreeShift, int i)
{
	octreeShift.levelOffset=1;
	calcRelativeOperation(root->operations[0], root->operations[i], octreeShift.relativeOps);
	octreeShift.indexOffset=root->infoStruct.nodeCount*i +1;
	octreeShift.vPageOffset=root->infoStruct.totalVPageCount*i;
	octreeShift.fPageOffset=root->infoStruct.totalFPageCount*i;
}

void OctreeLoader::calcNewOctreeRoot(Octree& newRoot)
{
	memcpy(&newRoot, &root[0], sizeof(Octree));
	reverseCalcCenter(root[0].center, root[0].width, root->operations[0], newRoot.center);
	//in order to make data dense, we use half width. maybe 1/4?
	//reverseCalcCenter(root[0].center, root[0].width/4.0, root->operations[0], newRoot.center);
	newRoot.width=root[0].width*2.0;
	newRoot.vNumber=root[0].vNumber*8;
	newRoot.fNumber=root[0].fNumber*8;
	newRoot.vRepeatNumber=root[0].vRepeatNumber*8;
	newRoot.fRepeatNumber=root[0].fRepeatNumber*8;
	newRoot.parentIndex=-1;
	newRoot.sonCount=8;
	newRoot.level=0;
	newRoot.myNodeIndex=0;
	newRoot.vAddress=newRoot.fAddress=NULL;
	newRoot.vFileAddLow=newRoot.vFileAddHigh=newRoot.fFileAddLow=newRoot.fFileAddHigh=-1;

	for (int i=0; i<8; i++)
	{
		newRoot.sons[i]=root->infoStruct.nodeCount*i+1;
	}
}
	

void OctreeLoader::expandBy8()
{
	Octree* octreeBuffer;
	Octree newRoot;

	OctreeShiftDescriptor octreeShift;
	int i, j;

	initialize(false);
	///////////////////////////////////////////
	//let's just set vertex file and face file to 8 time big as old one
	LONGLONG newSize;
	unsigned long int newSizeHigh=0, newSizeLow=0;

	//printf("totalpage=%u\n", root->infoStruct.totalVPageCount);
	
	newSize=(LONGLONG)(root->infoStruct.totalVPageCount*8)*(LONGLONG)(PageSize);
	//printf("newsize=%u\n", newSize);
	newSizeLow=newSize;
	newSizeHigh=newSize>>32;
	//printf("set file pointer high=%u, low=%u\n", newSizeHigh, newSizeLow);
	GenericSetFilePointer(root->vertexHandle, newSizeLow, newSizeHigh);
	GenericWriteFile(root->vertexHandle, &newSize, 1);

	newSize=(LONGLONG)(root->infoStruct.totalFPageCount*8)*(LONGLONG)PageSize;
	newSizeLow=newSize;
	newSizeHigh=newSize>>32;
	GenericSetFilePointer(root->faceHandle, newSizeLow, newSizeHigh);
	GenericWriteFile(root->faceHandle, &newSize, 1);
	///////////////////////////////////////////


	octreeBuffer=new Octree[root->infoStruct.nodeCount];
	CHECK_NEW(octreeBuffer);
	//considering the original cube is the first, we only need shift 7
	
	calcNewOctreeRoot(newRoot);
	vectorAssign(newRoot.center, octreeShift.newCenter);
	
	GenericSetFilePointer(root->octreeHandle, 0, 0);
	GenericWriteFile(root->octreeHandle, &newRoot, sizeof(Octree));
	octreeShift.vertexOffset=root->width;	
	for (i=0; i<8; i++)
	{		
		printf("begin expand round %d...\n", i);
		calcOctreeShiftDescriptor(octreeShift, i);
		//printf("after calc octree shift descriptor\n");

		for (j=0; j<root->infoStruct.nodeCount; j++)
		{
			calcShiftedLeafNode(root+j, octreeShift, octreeBuffer+j);
			//printf("after calc shifted leaf node\n");
			if (i!=0&&root[j].sonCount==0)
			{
				//i==0 no need copy
				//printf("before copy octree leaf node\n");
				copyOctreeLeafNode(root+j, octreeBuffer+j, octreeShift);
				//printf("after copy octree leaf node \n");
			}
			if (j==0)
			{
				octreeBuffer[0].parentIndex=0;
			}
		}
		//printf("after copy octree leaf nodes...\n");
		GenericWriteFile(root->octreeHandle, octreeBuffer, sizeof(Octree)*root->infoStruct.nodeCount);
		printf("shift-expand finishes %d ...\n", i);
	}
	delete[]octreeBuffer;
	GenericCloseFileMapping(vertexMapHandle);
	GenericCloseFileMapping(faceMapHandle);

	root->infoStruct.leafNodeNumber*=8;
	root->infoStruct.nodeCount=root->infoStruct.nodeCount*8+1;
	memcpy(root->infoStruct.origin, root[0].center, sizeof(VVector));
	root->infoStruct.totalVPageCount*=8;
	root->infoStruct.totalFPageCount*=8;
	root->infoStruct.totalFace*=8;
	root->infoStruct.totalVertex*=8;
	root->infoStruct.totalNeighbourCount*=8;
	root->infoStruct.totalRepeatVertex*=8;
	root->infoStruct.totalRepeatFace*=8;
	root->infoStruct.maxDepth+=1;

	while (root->infoStruct.maxNodeArrayLength<=root->infoStruct.nodeCount)
	{
		root->infoStruct.maxNodeArrayLength+=NodeArrayLengthIncrement;
	}
	GenericSetFilePointer(root->infoHandle, 0, 0);
	GenericWriteFile(root->infoHandle, &(root->infoStruct), sizeof(InfoStruct));

	delete[] Octree::root;
	GenericCloseFile(Octree::infoHandle);
	GenericCloseFile(Octree::octreeHandle);
	GenericCloseFile(Octree::vertexHandle);
	GenericCloseFile(Octree::faceHandle);
	//now let's update infoStruct

}

//end of expand by 8
////////////////////////////////////////////////////////////////////////////////////
	


void OctreeLoader::paintPseudoColor()
{
	int i, j;
	PlyVertex* vertexPtr;
	//int* intPtr;
	//int temp;
	//const unsigned int HeadMask=0x00FFFFFF;
	printf("starts creating pseudo-color data...\n");
	initialize(false);
	printf("finishes initializing...\n");
	for (i=0; i<root->infoStruct.nodeCount; i++)
	{
		if (root[i].sonCount==0)
		{
			/*
			if ((vertexPtr=(PlyVertex*)MapViewOfFile(vertexMapHandle, FILE_MAP_WRITE,
				root[i].vFileAddHigh, root[i].vFileAddLow, root[i].vMapSize))==0)
			{
				printf("map vertex failed %d\n", GetLastError());
				exit(4);
			}
			*/
			vertexPtr=(PlyVertex*)GenericMapViewOfFile(vertexMapHandle, false, 
				root[i].vFileAddHigh, root[i].vFileAddLow, root[i].vMapSize);

			for (j=0; j<root[i].vNumber+root[i].vRepeatNumber; j++)
			{				
				//this will have trouble for 64bit integer.
				memcpy(vertexPtr[j].color, &i, sizeof(UCVector));
				//intPtr=(int*)(vertexPtr[j].color);
				//temp=(*intPtr)&HeadMask;
			}
			/*
			if (UnmapViewOfFile(vertexPtr)==0)
			{
				printf("unmap failed %d\n", GetLastError());
				exit(34);
			}
			*/
			GenericUnmapViewOfFile(vertexPtr, root[i].vMapSize);
		}
	}
	printf("finishes creating pseudo color data...\n");
	uninitialize();
}


void OctreeLoader::readInfo()
{
	//Octree::readInfo(true);
	Octree::initialize(true);
	Octree::infoStruct.display();
}


void OctreeLoader::initialize(bool readOnly, bool fakeColor)
{
	//octree= new Octree;
	//DWORD accessRight=readOnly?PAGE_READONLY:PAGE_READWRITE;

	Octree::initialize(readOnly, fakeColor);
	root=Octree::root;
	vertexMapHandle=GenericCreateFileMapping(Octree::vertexHandle, readOnly);
	faceMapHandle=GenericCreateFileMapping(Octree::faceHandle, readOnly);
	/*
	vertexMapHandle=CreateFileMapping(octree->vertexHandle, NULL, accessRight, 0, 0, NULL);
	if (vertexMapHandle==NULL)
	{			
		printf("error code=%d\n", GetLastError());
		exit(2);
	}
	
	faceMapHandle=CreateFileMapping(octree->faceHandle, NULL, accessRight, 0, 0, NULL);
	if (faceMapHandle==NULL)
	{			
		printf("error code=%d\n", GetLastError());
		exit(2);
	}
	*/
}

void OctreeLoader::uninitialize()
{
	Octree::uninitialize(true);
	GenericCloseFileMapping(vertexMapHandle);
	GenericCloseFileMapping(faceMapHandle);
}

/*	
OctreeLoader::OctreeLoader()
{
	if ((mutex=CreateMutex(NULL, FALSE, NULL))==NULL)
	{
		printf("create mutex failed\n");
		exit(89);
	}

	if ((loadingThread=CreateThread(NULL, 0, loadingProc, NULL, 0, NULL))==NULL)
	{
		printf("create loading thread failed\n");
		exit(8);
	}

	//InitializeCriticalSection(&criticalSection);
	//printPriority();	
}
*/

void OctreeLoader::addCandidates(int count, int* array)
{
	for (int i=0; i<count; i++)
	{
		addCandidate(array[i]);
	}
}

/*
void OctreeLoader::calculateOctree(int nodeIndex)
{
	Octree* ptr=root+nodeIndex;
	int i;
	int result=frustum.cubeInsideFrustum(ptr->center[0], ptr->center[1], 
		ptr->center[2], ptr->width/2.0);

	switch (result)
	{
	case 1:
		//inside
		nodeSet[nextIndex].insert(nodeIndex);		
		break;
	case 0:
		//intersect
		if (ptr->sonCount==0)
		{
			nodeSet[nextIndex].insert(nodeIndex);	
		}
		else
		{
			for (i=0; i<8; i++)
			{
				if (ptr->sons[i]!=-1)
				{
					calculateOctree(ptr->sons[i]);
				}
			}
		}
		break;
	case -1:
		//outside
		break;
	}
}
*/


/*
bool OctreeLoader::testAndAdd(int nodeIndex)
{
	Octree*ptr=root+nodeIndex;

	//we only counts the leaf node data
	faceCounter+=root[nodeIndex].fNumber;
	vertexCounter+=root[nodeIndex].vNumber;
	nodeSet[nextIndex].insert(nodeIndex);
	return true;
}
*/



/*
bool OctreeLoader::testAndAdd(int nodeIndex)
{
	int testing;
	Octree*ptr=root+nodeIndex;
	//insert successful
	testing=frustum.CubeInFrustum(ptr->center[0], ptr->center[1], ptr->center[2], 
		ptr->width/2.0);
	if (testing==0||testing==1)
	{    
		//we only counts the leaf node data
		faceCounter+=root[nodeIndex].fNumber;
		vertexCounter+=root[nodeIndex].vNumber;
		nodeSet[nextIndex].insert(nodeIndex);	
		return true;
	}   
	else
	{
		return false;
	}
}
*/

void OctreeLoader::addCandidate(int index)
{
	nodeSet[nextIndex].insert(index);

	mapOctreeNode(index);
/*
	if (WaitForSingleObject(mutex, INFINITE)!=WAIT_OBJECT_0)
	{
		printf("wait failed %d\n", GetLastError());
		exit(87);
	}
	loadingQueue.push_back(index);
	//mapOctreeNode(index);
	if (!ReleaseMutex(mutex))
	{
		printf("release error %d\n", GetLastError());
		exit(8);
	}
*/
}


//assume:node itself passed frustum testing, now put neighbours
//Please note here: it is possible the neighbour of leaf node is not leaf!!!!!


/*
void OctreeLoader::showConservative()
{
	int missNode=0, missVertex=0, missTriangle=0;
	int errorCount=0, i;
	int totalNode=0, totalVertex=0, totalTriangle=0;
	if (conservativeMode)
	{	
		for (i=0; i<root->infoStruct.nodeCount; i++)
		{
			if (root[i].sonCount==0)
			{					
				if (!frustum.canCullCube(root[i].center[0], root[i].center[1], 
					root[i].center[2], root[i].width/2.0))
				{  					
					totalNode++;
					totalVertex+=root[i].vNumber+root[i].vRepeatNumber;
					totalTriangle+=root[i].fNumber;
					if (nodeSet[nextIndex].find(i)==nodeSet[nextIndex].end())
					{
						missNode++;
						missVertex+=root[i].vNumber+root[i].vRepeatNumber;
						missTriangle+=root[i].fNumber;
					}
				}   
				else
				{
					if (nodeSet[nextIndex].find(i)!=nodeSet[nextIndex].end())
					{
						errorCount++;
					}
				}
			}
		}

		if (totalVertex!=0&&totalTriangle!=0&&nodeSet[nextIndex].size()!=0)
		{
			printf("missNode=%d[%f], \nmissVertex=%d[%f], \nmissTriangle=%d[%f], \
				\nerrorCount=%d\n", missNode, 100.0*(float)(missNode)/(float)totalNode, 
				missVertex, 100.0*(float)missVertex/(float)totalVertex, missTriangle, 
				100.0*(float)missTriangle/(float)totalTriangle, errorCount);
		}
	}
}
*/



void OctreeLoader::unmapUnusedNodes()
{
	OctreeNodeSet::iterator it, result, temp;
	bool canUnmap;
	if (loadedVertexCounter>capacityFactor*MaxVertexCapacity*ThresholdRate
		||loadedFaceCounter>capacityFactor*MaxFaceCapacity*ThresholdRate)
	{
		printf("begin unmap unused nodes...\n");
		it=loadedNodeSet.begin();
		
		while (it!=loadedNodeSet.end())	
		{
			canUnmap=true;
			
			for (int i=0; i<MaxCoherentFrameNumber; i++)
			{
				//printf("loaded size %d, nodeset size %d\n", loadedNodeSet.size(), nodeSet[i].size());
				if ((result=nodeSet[i].find(*it))!=nodeSet[i].end())
				{
					//we find it is used recently
					canUnmap=false;			
					//printf("find in %d set\n", i);
					//printf("source %d, target %d\n", *it, *result);
					break;
				}
			}
			if (canUnmap)
			{
				//unmap	we remove 
				/*
				if (UnmapViewOfFile(root[*it].vAddress)==0)
				{
					printf("unmap failed\n");
					exit(333);
				}
				*/
				GenericUnmapViewOfFile(root[*it].vAddress, root[*it].vMapSize);
				root[*it].vAddress=NULL;
				loadedVertexCounter-=(root[*it].vNumber+root[*it].vRepeatNumber);
				if (loadedVertexCounter==0)
				{
					printf("what?\n");
				}

				/*
				if (UnmapViewOfFile(root[*it].fAddress)==0)
				{
					printf("unmap failed\n");
					exit(322);
				}
				*/
				GenericUnmapViewOfFile(root[*it].fAddress, root[*it].fMapSize);

				root[*it].fAddress=NULL;
				loadedFaceCounter-=(root[*it].fNumber+root[*it].fRepeatNumber);

				//in windows, compiler don't complain. but g++ is a nusance
				//it=loadedNodeSet.erase(it);
				//in linux we have to change to like this:
				temp=it;
				it++;
				loadedNodeSet.erase(temp);
			}
			else
			{
				++it;
			}			
		}	
	}
}

/*
void OctreeLoader::printPriority()
{
	int priority;

	if ((priority=GetThreadPriority(loadingThread))==THREAD_PRIORITY_ERROR_RETURN)
	{
		printf("get priority failed\n");
		exit(8);
	}		

	switch (priority)
	{
	case THREAD_PRIORITY_ABOVE_NORMAL:
		printf("THREAD_PRIORITY_ABOVE_NORMAL\n");
		break;
	case THREAD_PRIORITY_BELOW_NORMAL:
		printf("THREAD_PRIORITY_BELOW_NORMAL\n");
		break;
	case THREAD_PRIORITY_HIGHEST:
		printf("THREAD_PRIORITY_HIGHEST\n");
		break;
	case THREAD_PRIORITY_IDLE:
		printf("THREAD_PRIORITY_IDLE\n");
		break;
	case THREAD_PRIORITY_LOWEST:
		printf("THREAD_PRIORITY_LOWEST\n");
		break;
	case THREAD_PRIORITY_NORMAL:
		printf("THREAD_PRIORITY_NORMAL\n");
		break;
	case THREAD_PRIORITY_TIME_CRITICAL:
		printf("THREAD_PRIORITY_TIME_CRITICAL\n");
		break;
	}	
}
*/


/*
	if (fullRenderMode)
	{
		for (i=0; i<root->infoStruct.nodeCount; i++)
		{
			if (root[i].sonCount==0)
			{	
				glEnableClientState (GL_VERTEX_ARRAY);
				glEnableClientState (GL_NORMAL_ARRAY);
				glEnableClientState (GL_COLOR_ARRAY);
				mapOctreeNode(i);
				displayOctreeLeafNode(i);
				glDisableClientState (GL_VERTEX_ARRAY);
				glDisableClientState (GL_NORMAL_ARRAY);
				glDisableClientState (GL_COLOR_ARRAY);
				/*
				if (!frustum.canCullCube(root[i].center[0], root[i].center[1], 
					root[i].center[2], root[i].width/2.0))
				{  
					glEnableClientState (GL_VERTEX_ARRAY);
					glEnableClientState (GL_NORMAL_ARRAY);
					glEnableClientState (GL_COLOR_ARRAY);
					mapOctreeNode(i);
					displayOctreeLeafNode(i);
					glDisableClientState (GL_VERTEX_ARRAY);
					glDisableClientState (GL_NORMAL_ARRAY);
					glDisableClientState (GL_COLOR_ARRAY);
					hitCount++;
				}
				else
				{
					cullCount++;
				}
			

			}
		}
		//printf("cullCount=%d, hitCount=%d\n", cullCount, hitCount);
		return;
	}
	*/

void OctreeLoader::swapBuffer()
{
	//int i, cullCount=0, hitCount=0;
	int i;
	OctreeNodeSet::iterator it;
	//printPriority();
	
	unmapUnusedNodes();	
	/*
	//this must be safe
	if (WaitForSingleObject(mutex, INFINITE)!=WAIT_OBJECT_0)
	{
		printf("wait failed %d\n", GetLastError());
		exit(67);
	}
	unmapUnusedNodes();	
	if (!ReleaseMutex(mutex))
	{
		printf("release error %d\n", GetLastError());
		exit(8);
	}
	*/
	//loadingQueue.clear();
	
	//waitLoadingThread();

	//showConservative();

	//printf("currentset size=%d and nextset size=%d\n", nodeSet[currentIndex].size(), nodeSet[nextIndex].size());

	//we will unmap only when both previous and current node set is no longer needed


	glEnableClientState (GL_VERTEX_ARRAY);
	glEnableClientState (GL_NORMAL_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);
	//printf("nextset size=%d\n", nodeSet[nextIndex].size());
	for (it=nodeSet[nextIndex].begin(); it!=nodeSet[nextIndex].end(); it++)
	{			
		displayOctreeLeafNode(*it);			
	}
	glDisableClientState (GL_VERTEX_ARRAY);
	glDisableClientState (GL_NORMAL_ARRAY);
	glDisableClientState (GL_COLOR_ARRAY);

	currentIndex=nextIndex;
	nextIndex=(nextIndex+1)%MaxCoherentFrameNumber;
	nodeSet[nextIndex].clear();
	
	//Sleep(0);
}



/*
void OctreeLoader::waitLoadingThread()
{
	while (true)
	{
		if (WaitForSingleObject(mutex, INFINITE)!=WAIT_OBJECT_0)
		{
			printf("wait event failed\n");
			exit(898);
		}

		if (loadingQueue.empty())
		{
			if (!ReleaseMutex(mutex))
			{
				printf("release mutex failed %d\n", GetLastError());
				exit(1);
			}
			break;
		}

		if (!ReleaseMutex(mutex))
		{
			printf("releate mutex failed\n");
			exit(GetLastError());
		}
	}
}
*/


void OctreeLoader::mapOctreeNode(int index)
{
	//EnterCriticalSection(&criticalSection);
	
	if (root[index].vAddress==NULL)
	{		
		/*
		if ((root[index].vAddress=MapViewOfFile(vertexMapHandle, FILE_MAP_READ, 
			root[index].vFileAddHigh, root[index].vFileAddLow, root[index].vMapSize))==0)
		{
			printf("map vertex failed %d\n", GetLastError());
			exit(4);
		}
		*/
		root[index].vAddress=GenericMapViewOfFile(vertexMapHandle, true, root[index].vFileAddHigh, 
			root[index].vFileAddLow, root[index].vMapSize);

		loadedVertexCounter+=root[index].vNumber+root[index].vRepeatNumber;
		loadedNodeSet.insert(index);			
	}
	/////////////////////////////

	//////////////////////////////

	if (root[index].fAddress==NULL)
	{
		/*
		if((root[index].fAddress=MapViewOfFile(faceMapHandle, FILE_MAP_READ, 
			root[index].fFileAddHigh, root[index].fFileAddLow, root[index].fMapSize))==0)
		{
			printf("map face failed %d\n", GetLastError());
			exit(4);
		}
		*/
		root[index].fAddress=GenericMapViewOfFile(faceMapHandle, true, root[index].fFileAddHigh, 
			root[index].fFileAddLow, root[index].fMapSize);

		loadedFaceCounter+=root[index].fNumber+root[index].fRepeatNumber;		
	}
	//LeaveCriticalSection(&criticalSection);
	/*
	if (ReleaseMutex(mutex)==0)
	{
		printf("release failed=%d\n", GetLastError());
		exit(78);
	}
	*/
	
	/////////////////////////////
}



/*
OctreeLoader::~OctreeLoader()
{
	canTerminate=true;
	ReleaseMutex(mutex);
}
*/
void outputTriangles(char* fileName, int number, PlyVertex* vPtr, UIVector* fPtr)
{
	FILE* stream;
	stream=fopen(fileName, "w");
	for (int i=0; i<number; i++)
	{
		fprintf(stream, "no#%d triangle:\n", i);
		for (int j=0; j<3; j++)
		{
			fprintf(stream, "[%f,%f,%f],", vPtr[fPtr[i][j]].coord[0], 
				vPtr[fPtr[i][j]].coord[1], vPtr[fPtr[i][j]].coord[2]);
		}
		fprintf(stream, "\n");
	}
	fclose(stream);
}


void OctreeLoader::displayOctreeLeafNode(int index)
{
	PlyVertex* vPtr;
	UIVector* fPtr;
	//static bool doOnce=true;
	

	//mapOctreeNode(index);
	if (root[index].vAddress==NULL ||root[index].fAddress==NULL)
	{
		printf("unexpected error\n");
		exit(87);
	}
	
	vPtr=(PlyVertex*)(root[index].vAddress);
	fPtr=(UIVector*)(root[index].fAddress);
	

	getLastError();
#ifdef USE_DOUBLE_PRECISION
	glVertexPointer(3, GL_DOUBLE, sizeof(PlyVertex), &(vPtr->coord));
#else
	glVertexPointer(3, GL_FLOAT, sizeof(PlyVertex), &(vPtr->coord));
#endif

	glNormalPointer(GL_FLOAT, sizeof(PlyVertex), &(vPtr->normal));
	glColorPointer(3, GL_UNSIGNED_BYTE, sizeof(PlyVertex), &(vPtr->color));
	getLastError();

	glDrawElements(GL_TRIANGLES, (root[index].fNumber+root[index].fRepeatNumber)*3, 
		GL_UNSIGNED_INT, fPtr);

	getLastError();
}


bool OctreeLoader::displayOctreeLeafNodeRepeat(int index, int approach)
{
	PlyVertex* vPtr;
	UIVector* fPtr;
	int startingIndex, faceNumber;
	//mapOctreeNode(index);
	if (root[index].vAddress==NULL ||root[index].fAddress==NULL)
	{
		printf("unexpected error\n");
		exit(87);
	}
	
	if (approach*DisplayOctreeLeafNodeBatchNumber>=root[index].fNumber)
	{
		return true;
	}
	else
	{
		startingIndex = approach * DisplayOctreeLeafNodeBatchNumber;
		faceNumber = root[index].fNumber - startingIndex;
		faceNumber = MIN(faceNumber, DisplayOctreeLeafNodeBatchNumber);
		vPtr=(PlyVertex*)(root[index].vAddress);
		fPtr=(UIVector*)(root[index].fAddress);
		fPtr += startingIndex;
	}
	

	getLastError();
#ifdef USE_DOUBLE_PRECISION
	glVertexPointer(3, GL_DOUBLE, sizeof(PlyVertex), &(vPtr->coord));
#else
	glVertexPointer(3, GL_FLOAT, sizeof(PlyVertex), &(vPtr->coord));
#endif

	glNormalPointer(GL_FLOAT, sizeof(PlyVertex), &(vPtr->normal));
	glColorPointer(3, GL_UNSIGNED_BYTE, sizeof(PlyVertex), &(vPtr->color));
	getLastError();

	glDrawElements(GL_TRIANGLES, faceNumber*3, GL_UNSIGNED_INT, fPtr);

	getLastError();
	return false;
}


void getLastError()
{
	GLenum err;
	char msg[256];
	err=glGetError();
	switch(err)
	{
	case GL_NO_ERROR:
		sprintf(msg, "GL_NO_ERROR");
		break;
	case GL_INVALID_ENUM:
		sprintf(msg, "GL_INVALID_ENUM");
		break;
	case GL_INVALID_VALUE:
		sprintf(msg, "GL_INVALID_VALUE");
		break;
	case GL_INVALID_OPERATION:
		sprintf(msg, "GL_INVALID_OPERATION");
		break;
	case GL_STACK_OVERFLOW:
		sprintf(msg, "GL_STACK_OVERFLOW");
		break;
	case GL_STACK_UNDERFLOW:
		sprintf(msg, "GL_STACK_UNDERFLOW");
		break;
	case GL_OUT_OF_MEMORY:
		sprintf(msg, "GL_OUT_OF_MEMORY");
		break;
	}
	if (err!=GL_NO_ERROR)
	{
		printf("error message %s\n", msg);
		exit(err);
	}
}

