#include <string>
#include <cstdio>
#include <vector>
#include "rply.h"
#include <cmath>
#include "plyModel.h"
#include "octree.h"
//#include "plyloader.h"

#ifndef LINUX_VERSION
#include <windows.h>
#include <winbase.h>
#include "except.h"
#endif


using namespace std;

PlyModel ply;


string fileName;
VertexVector vertexVector;
FaceVector faceVector;
FaceStruct face;


static int callback(p_ply_argument argument)
{
    void *pdata;
    /* just pass the value from the input file to the output file */
    ply_get_argument_user_data(argument, &pdata, NULL);
    ply_write((p_ply) pdata, ply_get_argument_value(argument));
    return 1;
}

static int setup_callbacks(p_ply iply, p_ply oply) 
{
    p_ply_element element = NULL;
    /* iterate over all elements in input file */
    while ((element = ply_get_next_element(iply, element))) {
        p_ply_property property = NULL;
        long nelems = 0;
        const char *element_name;
        ply_get_element_info(element, &element_name, &nelems);
        /* add this element to output file */
        if (!ply_add_element(oply, element_name, nelems)) return 0;
        /* iterate over all properties of current element */
        while ((property = ply_get_next_property(element, property))) {
            const char *property_name;
            e_ply_type type, length_type, value_type;
            ply_get_property_info(property, &property_name, &type, 
                    &length_type, &value_type);
            /* setup input callback for this property */
            if (!ply_set_read_cb(iply, element_name, property_name, callback, 
                    oply, 0)) return 0;
            /* add this property to output file */
            if (!ply_add_property(oply, property_name, type, length_type, 
                    value_type)) return 0;
        }
    }
    return 1;
}


int convert(char* fileName)
{
    const char *value;
	p_ply iply, oply; 

    iply = ply_open(fileName, NULL);
    if (!iply) return 1; 
    if (!ply_read_header(iply)) return 1; 	

    oply = ply_create(fileName, PLY_LITTLE_ENDIAN, NULL);
    if (!oply) return 1;
    if (!setup_callbacks(iply, oply)) return 1; 
    //pass comments and obj_infos from input to output 
    value = NULL;
    while ((value = ply_get_next_comment(iply, value)))
        if (!ply_add_comment(oply, value)) return 1; 
    value = NULL;
    while ((value = ply_get_next_obj_info(iply, value)))
        if (!ply_add_obj_info(oply, value)) return 1;;
    // write output header
    if (!ply_write_header(oply)) return 1; 
    // read input file generating callbacks that pass data to output file /
    if (!ply_read(iply)) return 1; 
    // close up, we are done 
    if (!ply_close(iply)) return 1; 
    if (!ply_close(oply)) return 1;

    return 0;
}




VType calcEdge(VVector v1, VVector v2)
{
	VType result=0;
	for (int i=0; i<3; i++)
	{
		result+=(v1[i]-v2[i])*(v1[i]-v2[i]);
	}
	return sqrt(result);
}

bool PlyModel::calcNormal(FaceStruct& faceStruct, VertexVector& vertexVector)
{
	VType temp;
	int i, j;
	VVector product, first, second;

	for (i=0; i<3; i++)
	{	
		temp=calcEdge(vertexVector[faceStruct[i]].coord, vertexVector[faceStruct[(i+1)%3]].coord);
		if (temp>maxEdge)
		{
			maxEdge=temp;
		}
		if (temp==0)
		{
			printf("data corrupted!! edge is 0!face[%d,%d,%d]\n", faceStruct[0], faceStruct[1], faceStruct[2]);
			for (j=0; j<3; j++)
			{
				printf("v1[%d]=%f,v2[%d]=%f\n", j,  vertexVector[faceStruct[i]].coord[j], j, vertexVector[faceStruct[(i+1)%3]].coord[j]);
			}
 			return false;
			//exit(1594);
		}
	}
	for (i=0; i<3; i++)
	{
		first[i]=vertexVector[faceStruct[1]].coord[i]-vertexVector[faceStruct[0]].coord[i];
		second[i]=vertexVector[faceStruct[2]].coord[i]-vertexVector[faceStruct[0]].coord[i];
	}

	crossProduct(first, second, product);
	//now calculate average normal for each vertex by using face normal as vertex normal
	for (i=0; i<3; i++)
	{
		//for each vertex of triangle
		for (j=0; j<3; j++)
		{
			//for x,y,z of normal
			vertexVector[faceStruct[i]].normal[j]+=product[j];
			vertexVector[faceStruct[i]].normal[j]/=2.0;
		}
		normalize(vertexVector[faceStruct[i]].normal);
	}
	return true;
}


void PlyModel::normalize(VVector vect)
{
	VType tempSum=0.0;
	int i;
	for (i=0; i<3; i++)
	{
		tempSum+=vect[i]*vect[i];
	}
	if (tempSum==0.0)
	{
		return;//no need for normalize
	}
	for (i=0; i<3; i++)
	{
		vect[i]/=tempSum;
	}
}

void PlyModel::crossProduct(VVector first, VVector second, VVector result)
{
	result[0]=first[1]*second[2]-second[1]*first[2];
	result[1]=first[2]*second[0]-first[0]*second[2];
	result[2]=first[0]*second[1]-first[1]*second[0];
}


void PlyModel::calcLongestEdge(UIVector tri)
{
	VType result=0, temp;
	//float edge[3];
	//int first, second;
	for (int i=0; i<3; i++)
	{		
		temp=calcEdge(vertex[tri[i]].coord, vertex[tri[(i+1)%3]].coord);
		if (temp>maxEdge)
		{
			maxEdge=temp;
		}
	}
}


bool PlyModel::isValidTriangle(UIVector& myFace)
{
	VType result=0, temp;

	for (int i=0; i<3; i++)
	{		
		temp=calcEdge(vertex[myFace[i]].coord, vertex[myFace[(i+1)%3]].coord);
		if (result<temp)
		{
			result=temp;
		}
	}
	return result>MinimumEdgeLength;	
}


/*
void PlyModel::getVertexFromIndex(HANDLE vertexHandle, int index, PlyVertex& plyVertex)
{
	LONGLONG li;
	long int lo, hi;
	unsigned long int length;
	li=Int32x32To64(index, sizeof(PlyVertex));
	lo=li;
	hi=li>>32;
	if (SetFilePointer(vertexHandle, lo, &hi, FILE_BEGIN)==0xFFFFFFFF)
	{
		printf("move file pointer failed\n");
		exit(8);
	}
	if (ReadFile(vertexHandle, &plyVertex, sizeof(PlyVertex), &length, NULL)==0)
	{
		printf("read vertex failed\n");
		exit(12);
	}
}
*/


void PlyModel::getTriangleCenter(PlyVertex triangle[3], PlyVertex& center)
{
	int i, j;
	//float totalNormal=0;
	int tempColor;
	for (i=0; i<3; i++)
	{
		center.color[i]=0;
		center.coord[i]=0;
		center.normal[i]=0;
		tempColor=0;
		for (j=0; j<3; j++)
		{
			center.coord[i]+=triangle[j].coord[i];
			//center.color[i]+=triangle[j].color[i];
			tempColor+=triangle[j].color[i];//because it WILL overflow
			center.normal[i]+=triangle[j].normal[i];
		}
		center.color[i]=tempColor/3;
		center.coord[i]/=3.0;
		center.normal[i]/=3.0;//not normalized!!!
		//totalNormal+=center.normal[i]*center.normal[i];
	}
	/*
	totalNormal=sqrtf(totalNormal);
	for (i=0; i<3; i++)
	{
		center.normal[i]/=totalNormal;
	}
	*/

}


void PlyModel::divideFileByFour(HANDLE vertexHandles[4], HANDLE faceHandles[4])
{
	int i, j, k, mySize;

	int triangleCount[4]={0};

	UIVector myNewFace;
	UIVector* myFace;
	PlyVertex* myVertex;


	expandModel();

	mySize=fNumber/4;


	//for every group
	for (i=0; i<mySize; i++)
	{
		//for every partition
		for (j=0; j<4; j++)
		{
			myFace=triangle+i*4+j;

			//here let's check if the triangle is a valid one
			if (isValidTriangle(*myFace))
			{
				//then write the triangle and face into its file
				for (k=0; k<3; k++)
				{
					myNewFace[k]=triangleCount[j]*3+k;
					myVertex=vertex+(*myFace)[k];
					GenericWriteFile(vertexHandles[j], myVertex, sizeof(PlyVertex));
				}
				GenericWriteFile(faceHandles[j], &myNewFace, sizeof(UIVector));

				triangleCount[j]++;
			}
		}
	}
	delete[]vertex;
	delete[]triangle;
}

void PlyModel::addToOctree(char* dirName, int expandNumber)
{
	vector<string> fileNameVector;
	GenericRetrieveAllFile(dirName, fileNameVector);

	for (int i=0; i<fileNameVector.size(); i++)
	{		
		printf("begin add file %s to octree...\n", fileNameVector[i].c_str());
		doAddToOctree(fileNameVector[i].c_str(), expandNumber);
	}

}

void PlyModel::addToOctree(char* dirName, VType width, int expandNumber)
{
	vector<string> fileNameVector;
	GenericRetrieveAllFile(dirName, fileNameVector);

	for (int i=0; i<fileNameVector.size(); i++)
	{		
		printf("begin add file %s to octree...\n", fileNameVector[i].c_str());
		doAddToOctree(fileNameVector[i].c_str(), width, expandNumber);
	}
}

VType plyOperations[MaxExpandNumber][3]=
{
	//{1,0,1},
	{1,0,0},
	{1,0,-1},
	{0,0,1},
	{0,0,0},
	{0,0,-1},
	{-1,0,1},
	{-1,0,0},
	{-1,0,-1}
};

void PlyModel::addToOctree(const char* singleFileName)
{
union Endian 
{
        float f;        
        int i;        
        unsigned char c[4];        
};
	int badTriangle=0;
	FILE* file;
	union Endian endian;
        char buffer[256];
	VertexVector vertexVector;
	FaceVector faceVector;
        int i, j;
        unsigned char cNumber;
	PlyVertex plyVertex;
	FaceStruct faceStruct;
	unsigned char swap;

	if ((file=fopen(singleFileName, "r+b"))==NULL)
	{
		printf("fopen failed \n");
		exit(34);
	}

	while (true)
	{
		fgets(buffer, 256, file);
		
		//vNumber and fNumber is set here
		if (analysis(buffer))
		{
			break;
		}
	}
		

	for (i=0; i<vNumber; i++)
	{
		for (j=0; j<3; j++)
		{
			//changing big Endian to little endian
			if (fread(endian.c, sizeof(float), 1, file)!=1)
			{
				printf("reading vertex failed\n");
				exit(343);
			}
			swap=endian.c[0];
			endian.c[0]=endian.c[3];
			endian.c[3]=swap;
			swap=endian.c[1];
			endian.c[1]=endian.c[2];
			endian.c[2]=swap;
			plyVertex.coord[j]=endian.f;
			//printf("v[%d][%d]=%f\n", i, j, plyVertex.coord[j]);
			//initialize color and normal			
			plyVertex.color[j]=128;
			plyVertex.normal[j]=0;
		}
		plyVertex.color[3]=255;
		vertexVector.push_back(plyVertex);
		if (i%100000==0)
		{
			printf("processing vertex of %d...\n", i);
		}
			
	}

	printf("finished vertex processing and size of vertex is %d\n", vertexVector.size());
	for (i=0; i<fNumber; i++)
	{
		if (fread(&cNumber, sizeof(unsigned char), 1, file)!=1)
		{
			printf("reading color failed\n");;
			exit(12);
		}
		if (cNumber!=3)
		{
			printf("found trouble, some polygon is not triangle\n");
			exit(14);
		}
		for (j=0; j<3; j++)
		{
			if (fread(endian.c, sizeof(int), 1, file)!=1)
			{
				printf("reading face index failed\n");;
				exit(12);
			}
			//converting to little endian
			swap=endian.c[0];
			endian.c[0]=endian.c[3];
			endian.c[3]=swap;
			swap=endian.c[1];
			endian.c[1]=endian.c[2];
			endian.c[2]=swap;
			faceStruct[j]=endian.i;
		}

		//calcLongestEdge(faceStruct);
		//longest edge and normal will be calculated together
		if (calcNormal(faceStruct, vertexVector))
		{
			faceVector.push_back(faceStruct);
		}
		else
		{
			//else we just ignore this triangle since it is corrupted
			badTriangle++;
		}
		if (i%100000==0)
		{
			printf("processing triangle of %d...\n", i);
		}
	}
	fclose(file);
	printf("bad triangle is %d\n", badTriangle);
	Octree::addToOctree(vertexVector, faceVector);
}


void PlyModel::doAddToOctree(const char* fileName, VType width, int expandNumber)
{
	FaceStruct faceStruct;

	PlyVertex plyVertex;
	int startIndex, endIndex, expandCount;
	int i, j, k;

	VertexVector vertexVector;
	FaceVector faceVector;

	doReadFile(fileName);

	for (i=0; i<fNumber; i++)
	{
		memcpy(faceStruct.face, triangle[i], sizeof(UIVector));
		faceVector.push_back(faceStruct);
	}

	if (expandNumber>=MaxExpandNumber)
	{
		expandNumber=MaxExpandNumber;
	}
	for (i=0; i<expandNumber; i++)
	{	
		vertexVector.clear();
		for (j=0; j<vNumber; j++)
		{
			memcpy(&plyVertex, vertex+j, sizeof(PlyVertex));
			plyVertex.coord[0]+=width*plyOperations[i][0];
			plyVertex.coord[1]+=width*plyOperations[i][1];
			plyVertex.coord[2]+=width*plyOperations[i][2];
			vertexVector.push_back(plyVertex);		
		}
		Octree::addToOctree(vertexVector, faceVector);
	
	}
}



void PlyModel::doAddToOctree(const char* fileName, int expandNumber)
{
	FaceStruct faceStruct;
	string stringBuffer;
	HANDLE vertexHandles[4];
	HANDLE faceHandles[4];
	char charBuffer[MAX_TEMP_NAME_LENGTH];
	int startIndex, endIndex, expandCount;
	int i, j, k;
	vector<string> vertexNameVector;
	vector<string> faceNameVector;
	vector<HANDLE> vertexHandleVector;
	vector<HANDLE> faceHandleVector;

	VertexVector vertexVector;
	FaceVector faceVector;

	doReadFile(fileName);


	vertexHandles[0]=GenericOpenTempFile("temp", charBuffer);
	stringBuffer.assign(charBuffer);
	vertexNameVector.push_back(stringBuffer);
	vertexHandleVector.push_back(vertexHandles[0]);
	//GenericWriteFile(vertexHandles[0], vertex, sizeof(PlyVertex)*vNumber);

	faceHandles[0]=GenericOpenTempFile("temp", charBuffer);
	stringBuffer.assign(charBuffer);
	faceNameVector.push_back(stringBuffer);
	faceHandleVector.push_back(faceHandles[0]);

	//GenericWriteFile(faceHandles[0], triangle, sizeof(UIVector)*fNumber);

	startIndex=endIndex=0;//this is insane!!!
	for (i=0; i<expandNumber; i++)
	{		
		//when i==0, expandCount=1;	
		startIndex=endIndex;
		endIndex=vertexHandleVector.size();		

		for (j=startIndex; j<endIndex; j++)
		{
			//first let create four temp file for preparation
			for (k=0; k<4; k++)
			{
				vertexHandles[k]=GenericOpenTempFile("temp", charBuffer);
				stringBuffer.assign(charBuffer);
				vertexNameVector.push_back(stringBuffer);
				vertexHandleVector.push_back(vertexHandles[k]);

				faceHandles[k]=GenericOpenTempFile("temp", charBuffer);
				stringBuffer.assign(charBuffer);
				faceNameVector.push_back(stringBuffer);
				faceHandleVector.push_back(faceHandles[k]);
			}

			if (i!=0)
			{
				vertexVector.clear();
				faceVector.clear();
			
				Octree::readDataFromFile(vertexHandleVector[startIndex+j], vertexVector);
				Octree::readDataFromFile(faceHandleVector[startIndex+j], faceVector);

				

				vNumber=vertexVector.size();
				vertex=new PlyVertex[vNumber];
				CHECK_NEW(vertex);			
				memcpy(vertex, &(*vertexVector.begin()), sizeof(PlyVertex)*vNumber);


				fNumber=faceVector.size();
				triangle=new UIVector[fNumber];
				CHECK_NEW(triangle);
				memcpy(triangle, &(*faceVector.begin()), sizeof(UIVector)*fNumber);
			}

			divideFileByFour(vertexHandles, faceHandles);
			GenericCloseTempFile(vertexHandleVector[j], vertexNameVector[j].c_str());
			GenericCloseTempFile(faceHandleVector[j], faceNameVector[j].c_str());
		}
	}

	//we need minus one
	startIndex=endIndex;
	endIndex=vertexHandleVector.size();


	for (i=startIndex; i<endIndex; i++)
	{
		vertexVector.clear();
		faceVector.clear();
		if (expandNumber>0)
		{
			Octree::readDataFromFile(vertexHandleVector[i], vertexVector);
			Octree::readDataFromFile(faceHandleVector[i], faceVector);	
		}
		else
		{
			for (j=0; j<vNumber; j++)
			{
				vertexVector.push_back(vertex[j]);
			}
			for (j=0; j<fNumber; j++)
			{
				memcpy(faceStruct.face, triangle+j, sizeof(UIVector));
				faceVector.push_back(faceStruct);
			}
		}
		Octree::addToOctree(vertexVector, faceVector);
		GenericCloseTempFile(vertexHandleVector[i], vertexNameVector[i].c_str());
		GenericCloseTempFile(faceHandleVector[i], faceNameVector[i].c_str());
	}
/*
	for (i=0; i<vertexHandleVector.size(); i++)
	{
		GenericCloseTempFile(vertexHandleVector[i], vertexNameVector[i].c_str());
		GenericCloseTempFile(faceHandleVector[i], faceNameVector[i].c_str());
	}
	*/
}




/*
void PlyModel::retrievePlyData(char* dirName, Octree* octree, int expandNumber, bool toConvert)
{
	int i, j;


	GenericRetrieveAllFile(dirName, fileNameVector);

	for (i=0; i<fileNameVector.size(); i++)
	{
		fileName=fileNameVector[i];
		if (toConvert)
		{
			if (convertFile(fileName.c_str())!=0)
			{
				printf("convert file %s failed\n", fileName.c_str());
				exit(12);
			}
		}
		printf("before reading file %s...\n", fileName.c_str());

	
		readFile(fileName.c_str(), expandNumber);

	
		vertexVector.clear();
		faceVector.clear();
		for (j=0; j<vNumber; j++)
		{
			vertexVector.push_back(vertex[j]);
		}

		
		for (j=0; j<fNumber; j++)
		{
			sortTriangleIndex(triangle[j]);
			face=triangle[j];
			faceVector.push_back(face);
		}

		delete[]vertex;
		delete[]triangle;
		octree->addToOctree(vertexVector, faceVector);	
	}
}
*/

bool PlyModel::analysis(char* buf)
{
	char* token;
	token=strtok(buf, " \n");
	while (token!=NULL)
	{
		if (strcmp(token, "vertex")==0)
		{
			token=strtok(NULL, " \n");
			sscanf(token, "%d", &vNumber);
			break;
		}
		if (strcmp(token, "face")==0)
		{
			token=strtok(NULL, " \n");
			sscanf(token, "%d", &fNumber);
			break;
		}
		if (strcmp(token, "end_header" )==0)
		{
			return true;
		}
		token=strtok(NULL, " \n");		
	}
	return false;
}

int PlyModel::convertFile(const char* fileName)
{
	return convert(const_cast<char*>(fileName));
}

int PlyModel::doReadFile(const char* fileName)
{
	FILE* file;

	char buffer[256];
	int i, j;
	unsigned char cNumber;
	UCVector dummy;

	if ((file=fopen(fileName, "r+b"))==NULL)
	{
		printf("fopen failed \n");
		exit(34);
	}

	while (true)
	{
		fgets(buffer, 256, file);
		//i=fread(buffer, 1, 256, file);
		
		//vNumber and fNumber is set here
		if (analysis(buffer))
		{
			break;
		}
	}
	//let's do the expanding here
	
	vertex=new PlyVertex[vNumber];

		//each triangle will become four triangles
	triangle=new UIVector[fNumber];

	if (vertex==NULL||triangle==NULL)
	{
		printf("allocate memory fails\n");
		exit(13);
	}

	for (i=0; i<vNumber; i++)
	{
		for (j=0; j<3; j++)
		{
			float temp;
			fread(&temp, sizeof(float), 1, file);
			vertex[i].coord[j]=temp;
		}
		//fread(vertex[i].coord, sizeof(float), 3, file);
		fread(vertex[i].color, sizeof(unsigned char), 3, file);
		vertex[i].color[3]=255;
		fread(vertex[i].normal, sizeof(float), 3, file);		
	}

	for (i=0; i<fNumber; i++)
	{
		if (fread(&cNumber, sizeof(unsigned char), 1, file)!=1)
		{
			printf("reading color failed\n");;
			exit(12);
		}
	
		if (cNumber!=3)
		{
			printf("found trouble, some polygon is not triangle\n");
			exit(14);
		}
		if (fread(triangle[i], sizeof(int), cNumber, file)!=cNumber)
		{
			printf("reading face index failed\n");;
			exit(12);
		}
		calcLongestEdge(triangle[i]);
	
		//discard face color
		if (fread(dummy, 3, 1, file)!=1)
		{
			printf("reading dummy color failed\n");;
			exit(12);
		}
	}
	//////////////////////////////////////
	//now let's find vertex by triangle.

	//int tripleIndex[3];

	fclose(file);
	//EXCEPT_CALL(expandModel(expandNumber), fileName);
	return 0;
}

/*
void PlyModel::outputTriangle(int triangleIndex)
{
	for (int i=0; i<3; i++)
	{
		printf("\n*******************output vertex%d\n", triangle[triangleIndex][i]);
		for (int j=0; j<3; j++)
		{
			printf("[%f]", vertex[triangle[triangleIndex][i]].coord[j]);
		}
		printf("\n");
	}
	printf("****************************\n");
}


bool PlyModel::equalVertex(int index1, int index2)
{
	for (int i=0; i<3; i++)
	{
		if (vertex[index1].coord[i]!=vertex[index2].coord[i])
		{
			return false;
		}
	}
	return true;
}


bool PlyModel::validTriangle(int triangleIndex)
{
	for (int i=0; i<3; i++)
	{
		if (equalVertex(triangle[triangleIndex][i], triangle[triangleIndex][(i+1)%3]))
		{
			printf("ooops\n");
			exit(1);
			return false;

		}
	}
	return true;
}
*/


void PlyModel::expandModel()
{

	int i, j, n, k, l;
	int vertexSize, faceSize;
	//PlyVertex oldTripleVertex[3];
	//PlyVertex newTripleVertex[3];
	PlyVertex* oldTripleVertex[3];
	PlyVertex* newTripleVertex[3];
	UIVector newFace;
	UIVector tripleFace[3];

	PlyVertex* newVertexPtr;
	UIVector* newTrianglePtr;

	vertexSize=vNumber+3*fNumber;
	faceSize=fNumber+3*fNumber;

	newVertexPtr=new PlyVertex[vertexSize];
	newTrianglePtr=new UIVector[faceSize];

	CHECK_NEW(newVertexPtr);
	CHECK_NEW(newTrianglePtr);

	//reallocate
	memcpy(newVertexPtr, vertex, sizeof(PlyVertex)*vNumber);
	memcpy(newTrianglePtr, triangle, sizeof(UIVector)*fNumber);
	delete[]vertex;
	delete[]triangle;
	vertex=newVertexPtr;
	triangle=newTrianglePtr;

	for (i=0; i<fNumber; i++)
	{
/////////////////////////////////////////////////////////////
		//printf("original vertex is \n");
		//outputTriangle(i);
	

		for (j=0; j<3; j++)
		{
			//memcpy(oldTripleVertex+j, vertex+triangle[i][j], sizeof(PlyVertex));
			oldTripleVertex[j]=vertex+triangle[i][j];
			newTripleVertex[j]=vertex+vNumber+i*3+j;
			//for every triangle, we increase three new vertex, their index starts after previous
			//newly-added triple vertex of previous triangle, all after original vertex index, vNumber
			newFace[j]=vNumber+i*3+j;//the three new vertex index is just added from end of old 
		}
		
		//tripleFace=triangle+fNumber+expandFactor*i;
		for (j=0; j<3; j++)
		{
			getMiddleVertex(oldTripleVertex[j], oldTripleVertex[(j+1)%3], newTripleVertex[j]);
			//memcpy(vertex+vNumber+i*3+j, newTripleVertex+j, sizeof(PlyVertex));				
		}

		for (j=0; j<3; j++)
		{
			tripleFace[j][0]=triangle[i][j];
			tripleFace[j][1]=newFace[j];
			//this is tricky!!!just think the position of new triangle
			tripleFace[j][2]=newFace[(j+2)%3];
		}

		for (j=0; j<3; j++)
		{
			calcLongestEdge(tripleFace[j]);
		}
		calcLongestEdge(newFace);
	
		//memcpy(triangle[i], tripleFace[0], sizeof(UIVector));
		//memcpy(triangle[i], newFace, sizeof(UIVector));//replace the old triangle
	
		memcpy(triangle[fNumber+i*3], tripleFace, sizeof(UIVector)*3);//replace the old triangle
		memcpy(triangle[i], newFace, sizeof(UIVector));//replace the old triangle
	}
	vNumber=vertexSize;//now updates the vertex number, because each face will generate a new triangle.
	//fNumber=4*fNumber;//each triangle will become three triangles
	fNumber=faceSize;
}


	

void PlyModel::getMiddleVertex(PlyVertex* old1, PlyVertex* old2, PlyVertex* result)
{
	int i, temp;
	float total=0.0;
	for (i=0; i<3; i++)
	{
		result->coord[i]=(old1->coord[i]+old2->coord[i])/2.0;
		temp=((int)(old1->color[i])+(int)(old2->color[i]))/2;
		result->color[i]=temp;
		result->normal[i]=(old1->normal[i]+old2->normal[i])/2.0;
		total+=result->normal[i]*result->normal[i];
	}
	result->color[3]=((int)(old1->color[3])+(int)(old2->color[3]))/2;
	total=sqrtf(total);
	for (i=0; i<3; i++)
	{
		result->normal[i]/=total;
	}
}

VVector tetrahedron[4]=
{
	{+1, +1, +1}, 
	{-1, -1, +1},
	{-1, +1, -1}, 
	{+1, -1, -1}
};

FaceStruct faceStructArray[6]=
{
	{0,1,2},
	{0,1,3},
	{0,2,3},
	{1,2,3}
};




void myTestData()
{
	VertexVector vertexVector;
	FaceVector faceVector;
	//FaceStruct faceStruct;
	PlyVertex plyVertex;
	int i, j;

	DWORD timeCount;

	timeCount=GenericGetTime();
	Octree::initialize(false);

	for (i=0; i<4; i++)
	{
		memcpy(plyVertex.coord, tetrahedron[i], sizeof(VVector));
		memcpy(plyVertex.normal, tetrahedron[i], sizeof(VVector));
		for (j=0; j<4; j++)
		{
			plyVertex.color[j]=100;
			plyVertex.color[i]=255;
		}
		vertexVector.push_back(plyVertex);
	}
	for (i=0; i<4; i++)
	{
		faceVector.push_back(faceStructArray[i]);
	}

	Octree::addToOctree(vertexVector, faceVector);

	Octree::uninitialize(false);
	timeCount=GenericGetTime()- timeCount;
	printf("convert file process takes time %d seconds\n", timeCount);
}


void mergeData(char* dirName, float width, int expandNumber)
{
	DWORD timeCount;
	timeCount=GenericGetTime();
	Octree::initialize(false);
	ply.addToOctree(dirName, width, expandNumber);
	Octree::uninitialize(false);
	timeCount=GenericGetTime()- timeCount;
	printf("convert file process takes time %d seconds\n", timeCount);

}

void mergeSingleFile(char* singleFileName)
{
	DWORD timeCount;
	timeCount=GenericGetTime();
	Octree::initialize(false);
	ply.addToOctree(singleFileName);
	Octree::uninitialize(false);
	timeCount=GenericGetTime()- timeCount;
	printf("convert file process takes time %d seconds\n", timeCount);
}

void mergeData(char* dirName, int expandTimes)
{
	DWORD timeCount;
	timeCount=GenericGetTime();
	Octree::initialize(false);
	//retrievePlyData(dirName, octree, expandTimes, false);
	ply.addToOctree(dirName, expandTimes);
	Octree::uninitialize(false);
	timeCount=GenericGetTime()- timeCount;
	printf("convert file process takes time %d seconds\n", timeCount);
}

/*
void retrievePlyData(char* dirName, Octree* octree, int expandNumber, bool toConvert)
{
	//DWORD timeCount;
	//timeCount=GetTickCount();

	ply.addToOctree(dirName, octree, expandNumber);

	//loadPlyFile(octree, dirName);

	//myTestInput(octree);
	//timeCount=GetTickCount()- timeCount;
	//printf("convert file process takes time %d seconds\n", timeCount);
	
}
*/


/*
void convertPlyFile(char* dirName, int expandNumber, bool toConvert)
{
	DWORD timeCount;
	HANDLE handle;
	unsigned long length;
	ply.maxEdge=0.0;


	timeCount=GetTickCount();
	ply.readDirectory(dirName, expandNumber, toConvert);

	timeCount=GetTickCount()- timeCount;
	printf("convert file process takes time %d seconds\n", timeCount/1000);
	
	handle=CreateFile("info.data", GENERIC_READ|GENERIC_WRITE, 0, NULL, 
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (handle==INVALID_HANDLE_VALUE)
	{
		printf("open info file failed\n");
		exit(89);
	}

	if (WriteFile(handle, &ply, sizeof(PlyModel),&length, NULL)==0)
	{
		printf("error of writing info and only write %d bytes %d\n", length,  GetLastError());
		exit(88);
	}

	printf("vcount=%d, fcount=%d\n", ply.vertexCount, ply.faceCount);
	for(int i=0; i<3; i++)
	{
		printf("max[%c]=%f, min[%c]=%f\n", 'x'+i, ply.maxCoord[i], 'x'+i, ply.minCoord[i]);
	}
	printf("max edge is %f\n", ply.maxEdge);
	CloseHandle(handle);

}
*/
