/***********************************************************************************


************************************************************************************/

#include <GL/glut.h>
#include "plyModel.h"
#include "octree.h"
#include "octreeLoader.h"
#include "viewCollector.h"
#include "renderEngine.h"
#include <sys/stat.h>
#include <cstdio>
#include <cmath>

#ifndef LINUX_VERSION
#include <windows.h>
#endif
//#include "plyloader.h"

#define RELEASE_VERSION 1



typedef GLfloat Color[4];
const VType DefaultPitch=0;
const VType DefaultYaw=45;
const VType DefaultRoll=0;
const Color color[MaxLightNumber] = 
{
	{ 1.0, 0.0, 0.0 , 1.0},	// RED
	{ 0.0, 1.0, 0.0 , 1.0},	// GREEN	
	{ 0.0, 0.0, 1.0 , 1.0},	// BLUE
	{ 1.0, 1.0, 0.0 , 1.0},	// YEWLLOW
	{ 0.0, 1.0, 1.0 , 1.0},	// MAGENTA	
	{ 1.0, 0.0, 1.0 , 1.0},	// PINK
	{ 1.0, 1.0, 1.0 , 1.0},	// WHITE
	{ 0.0, 0.0, 0.0 , 1.0},	// BLACK

};
const GLfloat no_mat[] = { 0.0, 0.0, 0.0, 1.0 };
const GLfloat half_mat[] = { 1.0, 1.0, 1.0, 1.0 };
const GLfloat light_diffuse[] = {0.9, 0.9, 0.9, 1.0};	//{ 1.0, 1.0, 1.0, 1.0 };
const GLfloat light_specular[]={1.0, 1.0, 1.0, 1.0};

GLfloat light_position[MaxLightNumber][4] = 
{ 
	{0.0, 0.0, -5.0, 1.0 },
	{0.0, 15.0, 0.0, 1.0 },
	{0.0, -15.0, 0.0, 1.0 },
	
	{0.0, 21.0, 0.0, 1.0 },
	{0.0, -10.0, 0.0, 1.0 },
	{0.0, -14.0, 0.0, 1.0 },
	{0.0, -16.0, 0.0, 1.0 },
	{0.0, -11.0, 0.0, 1.0 }
};

GLfloat light_direction[MaxLightNumber][3] = 
{ 
	{0.0, 0.0, 1.0 },
	{0.0, -1.0, 0.0 },
	{0.0, -1.0, 0.0 },
	{0.0, -1.0, 0.0 },
	{0.0, 1.0, 0.0 },
	{0.0, 1.0, 0.0 },
	{0.0, 1.0, 0.0 },
	{0.0, 1.0, 0.0 }
};

const GLfloat fire[] = { 1.0f, 0.6f, 0.2f, 0.8f };

GLenum lightEnum[MaxLightNumber]=
{
	GL_LIGHT0, GL_LIGHT1, GL_LIGHT2,  GL_LIGHT3, GL_LIGHT4, GL_LIGHT5, 
		GL_LIGHT6,GL_LIGHT7
};


void Lighting::init()
{
	int i;

	lightEnabled=false;
	for (i=0; i<currentLightNumber; i++)
	{			
		glLightfv(lightEnum[i], GL_AMBIENT, no_mat);
		//glLightfv(lightEnum[i], GL_DIFFUSE, light_diffuse);
		glLightfv(lightEnum[i], GL_DIFFUSE, color[i]);
		//glLightfv(lightEnum[i], GL_SPECULAR, light_specular);
		glLightfv(lightEnum[i], GL_SPECULAR, color[i]);

		glLightf(lightEnum[i], GL_CONSTANT_ATTENUATION, 1.0);
		glLightf(lightEnum[i], GL_LINEAR_ATTENUATION, 0.0);
		glLightf(lightEnum[i], GL_QUADRATIC_ATTENUATION, 0.0);	
		
		//glLightfv(lightEnum[i], GL_POSITION, light_position[i]);
		glLightfv(lightEnum[i], GL_POSITION, light_position[i]);
		glLightfv(lightEnum[i], GL_SPOT_DIRECTION, light_direction[i]);
		glLightf(lightEnum[i], GL_SPOT_CUTOFF, 80.0);
	}
	
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, color[WHITE]);
	glMaterialfv(GL_FRONT, GL_SPECULAR, color[WHITE]);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 50.0);

	//glEnable(GL_LIGHT0);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	
	/*
	glEnable(GL_COLOR_MATERIAL);

	for (i=0; i<currentLightNumber; i++)
	{
		glEnable(lightEnum[i]);
	}
	glEnable(GL_LIGHTING);
	*/

}

void Lighting::update(const Movement& movement)
{
	float temp[4];
	for (int i=0; i<currentLightNumber; i++)
	{		
		//memcpy(temp, light_position[i], sizeof(float)*4);
		light_position[i][0]+=movement.x;
		light_position[i][1]+=movement.y;
		light_position[i][2]+=movement.z;
		glLightfv(lightEnum[i], GL_POSITION, light_position[i]);
		light_direction[i][0]=-(light_position[i][0]-movement.fLook_x);
        	light_direction[i][1]=-(light_position[i][1]-movement.fLook_y);
        	light_direction[i][2]=-(light_position[i][2]-movement.fLook_z);
        	glLightfv(lightEnum[i], GL_SPOT_DIRECTION, light_direction[i]);
	}

}

void Lighting::toggleLighting()
{
	int i;
	if (lightEnabled)
	{
		glDisable(GL_COLOR_MATERIAL);			
		glDisable(GL_LIGHTING);
		for (i=0; i<currentLightNumber; i++)
		{
			glDisable(lightEnum[i]);
		}
		printf("lighting is disabled\n");			
	}
	else
	{
		glEnable(GL_COLOR_MATERIAL);
		
		for (i=0; i<currentLightNumber; i++)
		{
			glEnable(lightEnum[i]);
		}
		glEnable(GL_LIGHTING);
		printf("lighting is enabled\n");			
	}
	lightEnabled=!lightEnabled;
}

	//simpleInit();
	//simpleInit();
	//glutMainLoop();	

void RenderEngine::init(OctreeLoader* octreeLoader)
{
	//createFiles();
	flatShading=true;
	wireFrame=false;


	glShadeModel(GL_FLAT);
	setupPerspective(octreeLoader);
	lighting.currentLightNumber=8;
	lighting.init();
	//
	glEnable(GL_DEPTH_TEST);
	updatePosition();
}


//////////////////////////////////////////////////////////
//from old chopper
void RenderEngine::keyboardCallback(unsigned char key, int x, int y)
						// This is the callback procedure for capturing OpenGL Keyboard events.
{
	const int BlockSize=100;
	float buffer[BlockSize*BlockSize];
	VType fx, fy, fz;
	Movement temp=movement;
	int i, j;	
	switch(key)
	{
/*
	case 'z': //zoom in
		renderContext.fovy -= 0.5;
		//fViewVol -= 0.2;
		break;
	case 'Z'://zoom out
		renderContext.fovy += 0.5;
		//fViewVol += 0.2;
		break;

	case 'F': //move camera position forward
	case 'f':
		if(abs(radius - 0.1) > 0.0)	
		{
			radius -= 0.1;
		}
		break;

	case 'b': //move camera position backward
	case 'B':
		radius += 0.1;	
		break;
*/

	case 'v':
	case 'V':
		if (wireFrame)
		{
			glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		}
		else
		{	
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL); 
		}
		wireFrame=!wireFrame;
		break;
	case 'c':
	case 'C':
		initializeMovement(startingPos);
		break;
	case 'X' :
	case 'x' :
		showMenu();
		break;

	case 'n':
	case 'N':
		if (flatShading)
		{
			glShadeModel(GL_FLAT);	
			printf("flat shading\n");
		}
		else
		{
			glShadeModel(GL_SMOOTH);
			printf("smooth shading\n");
		}
		flatShading=!flatShading;
		break;
	case 'p':
		if (movement.pitch>-45)
		{			
			movement.pitch-=angleOffset;			
		}
		break;
	case 'P':
		if (movement.pitch<45)
		{			
			movement.pitch+=angleOffset;		
		}
		break;
	case 'y':		
		movement.yaw+=angleOffset;		
		if (movement.yaw>360)
		{
			movement.yaw-=360;
		}	
		break;
	case 'Y':	
		movement.yaw-=angleOffset;		
		if (movement.yaw<-360)
		{
			movement.yaw+=360;
		}			
		break;
	case 'r':
		if (movement.roll>-45)
		{
			movement.roll-=angleOffset;
		}
		break;
	case 'R':
		if (movement.roll<45)
		{
			movement.roll+=angleOffset;
		}
		break;
	case 'U':
		temp.pitch-=90;
		calcCoord(positionOffset, temp, fx, fy, fz);
		movement.x+=fx;
		movement.y+=fy;
		movement.z+=fz;			
		break;
	case 'u':		
		temp.pitch+=90;
		calcCoord(positionOffset, temp, fx, fy, fz);
		movement.x+=fx;
		movement.y+=fy;
		movement.z+=fz;			
		break;
	case 'D':
	case 'd':
		movement.yaw-=angleOffset;
		if (movement.yaw<-360)
		{
			movement.yaw+=360;
		}			
		break;
	case 'a':
	case 'A':
		movement.yaw+=angleOffset;
		if (movement.yaw>360)
		{
			movement.yaw-=360;
		}			
		break;
	case 's':
	case 'S':
		if (movement.pitch>-45)
		{			
			movement.pitch-=angleOffset;
			
		}
		break;
	case 'w':
	case 'W':
		if (movement.pitch<45)
		{			
			movement.pitch+=angleOffset;
		
		}
		break;

//////////////////////////////////////////////////////////////////////////////
	case 'j':		
	case 'J':	

		movement.yaw+=angleOffset;
		
		if (movement.yaw>360)
		{
			movement.yaw-=360;
		}			
	
		break;

	case 'l':
	case 'L':
		movement.yaw-=angleOffset;
		
		if (movement.yaw<-360)
		{
			movement.yaw+=360;
		}			
		
		break;

	case 'i':		// Rotate center of projection up
	case 'I':
		//theta -= 0.5;
		calcCoord(movementOffset, movement, fx, fy, fz);
	
		movement.x+=fx;
		movement.y+=fy;
		movement.z+=fz;
		break;

	case 'k':	// Rotate center of projection down
	case 'K':
		//theta += 0.5;
		//movement.z-=1;
		calcCoord(-movementOffset, movement, fx, fy, fz);
		movement.x+=fx;
		movement.y+=fy;
		movement.z+=fz;	
		break;
////////////////////////////////////////////////////////////////
		///////////////////////////////////////////////////////////

	case 'e':
	case 'E':
		lighting.toggleLighting();
		break;
	case  'z':
		glReadPixels(200, 200, BlockSize, BlockSize, GL_DEPTH_COMPONENT, GL_FLOAT, buffer);
		for (i=0; i<BlockSize; i++)
		{
			for (j=0; j<BlockSize; j++)
			{
				printf("[%f]", buffer[i*BlockSize+j]);
			}
			printf("\n");
		}
		break;
		
	case 27:				//ESCAPE Code for exiting program.
		exit(0);
	}
	updatePosition();
	glutPostRedisplay();
}




//end of old chopper
///////////////////////////////////////////////


void RenderEngine::initializeMovement(VVector origin)
{
	int i, j;
	movement.x=origin[0];
	movement.y=origin[1];
	movement.z=origin[2];
	movement.pitch=DefaultPitch;
	movement.roll=DefaultRoll;
	movement.yaw=DefaultYaw;
	for (i=0; i<MaxLightNumber; i++)
	{
		light_position[i][0]=movement.x+Octree::operations[i][0]*LightPositionOffset;
		light_position[i][1]=movement.x+Octree::operations[i][1]*LightPositionOffset;
		light_position[i][2]=movement.x+Octree::operations[i][2]*LightPositionOffset;
	}			
	
}

///////////////////////////////////////////////////////////////////
/////////////simple testing

void RenderEngine::updatePosition()
{
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	movement.fLook_x=movement.x+
		sin(deg2rad(movement.yaw))*cos(deg2rad(movement.pitch));
	movement.fLook_y=movement.y+
		sin(deg2rad(movement.pitch))*cos(deg2rad(movement.roll)); 
	movement.fLook_z=movement.z+
		cos(deg2rad(movement.yaw))*cos(deg2rad(movement.pitch));

	movement.fUp_x = sin(deg2rad(movement.roll));
	movement.fUp_y = cos(deg2rad(movement.roll));
	movement.fUp_z = 0;

	if (lighting.lightEnabled)
	{
		lighting.update(movement);
	}

	//printf("x=%f, y=%f, z=%f\n", movement.x, movement.y, movement.z);
	//printf("yaw=%f, pitch=%f, roll=%f\n", movement.yaw, movement.pitch, movement.roll);
	gluLookAt(movement.x, movement.y, movement.z, movement.fLook_x, movement.fLook_y, movement.fLook_z, 
		movement.fUp_x, movement.fUp_y, movement.fUp_z);


/////////////////////////////////////////////////////////////
	drawingContext.center[0]=movement.x;
	drawingContext.center[1]=movement.y;
	drawingContext.center[2]=movement.z;
}


void RenderEngine::setupPerspective(OctreeLoader* octreeLoader)
{
	int i, j;

	//initLights();
	//memcpy(startingPos, octreeLoader->root[0].center, sizeof(VVector));
	memcpy(startingPos, octreeLoader->root[0].center, sizeof(VVector));
	
	for (i=0; i<3; i++)
	{
		startingPos[i]+=octreeLoader->root[0].width*StartingOffset[i];
	}
	

	initializeMovement(startingPos);
	/*
	for (i=0; i<currentLightNumber; i++)
	{
		for (j=0; j<3; j++)
		{
			light_position[i][j]+=octreeLoader->root[0].center[j];
		}
	}
	*/
	drawingContext.fovy=1;
	drawingContext.aspect=1000.0;
	drawingContext.zFar=drawingContext.zNear=octreeLoader->root[0].width*10;
	for (i=0; i<=octreeLoader->root[0].infoStruct.maxDepth; i++)
	{
		drawingContext.zNear/=2.0;
	}
	drawingContext.zNear*=0.05;
	//positionOffset=movementOffset=drawingContext.zNear *100;
	positionOffset=movementOffset=drawingContext.zNear*2;
	angleOffset=5.0;
	if (drawingContext.zNear==0)
	{
		printf("how can this be true?\n");
		drawingContext.zNear=DefaultNearPlane;
	}
	//drawingContext.zNear=0.1;
	//glDepthFunc(GL_LEQUAL);
	//glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	//glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	//glShadeModel(GL_SMOOTH);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(drawingContext.fovy*60.0f, (VType)scrW/(VType)scrH, 
		drawingContext.zNear, drawingContext.zFar);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}

void RenderEngine::showMenu(void)
{
	printf("\n\n ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////");
	printf("\n OpenGL PROGRAM FOR WALK THROUGH POWER PLANT\n");
	printf("\n April 26, 2007 \n");
	printf("////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////\n");
	printf("\n\n Operations: ");
	
	printf("\n Use c/C key to clear viewing parameters and return to default.");
	printf("\n Use y/Y key for a Yaw rotation.");
	printf("\n Use p/P key for a pitch rotation.");
	printf("\n Use U/u/D/d to move chopper upward/downward vertically");
	printf("\n Use n/N to toggle smooth and flat shading");
	printf("\n Use arrow key to move forward and rotate");
	printf("\n Use e/E to toggle lighting");
	printf("\n Use x/X to show this menu");

	printf("\n Use b/B to start parallel rendering mode");
	
	printf("\n Use esc key to exit gracefully from program.");
}

inline VType RenderEngine::deg2rad(VType deg)
{
	return deg/360*2*PI;
}

inline VType RenderEngine::rad2deg(VType rad)
{
	return rad/2.0/PI*360.0;
}



void RenderEngine::calcCoord(VType distance, const Movement& position, VType &x, VType &y, VType &z)
{
	x=distance*cos(deg2rad(position.pitch))*sin(deg2rad(position.yaw));
	y=distance*sin(deg2rad(position.pitch))*cos(deg2rad(position.roll));

	z=distance*cos(deg2rad(position.pitch))*cos(deg2rad(position.yaw));
}
////////////////////////////////////////////////////////////////////////////////


void Movement::display()
{
	printf("x=%f,y=%f,z=%f\n", x, y, z);
	printf("yaw=%f,pitch=%f,roll=%f\n", yaw, pitch, roll);
}


