#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

const int ProfileItemCount=9;
const int MaxNodeNumber=256;

#define DISPLAY_EXCEL 1

enum ProfileItem
{
        CM=0, CT=2, CR=5, CD=7, BMTRD=1, BTRD=4, BRD=6, BT=3,  ALL=8
};


static char* ProfileItemName[ProfileItemCount]=
{
        "computation of Master",
        "communication from Master to Tester, Render and Displayer",
        "computation of Tester",
        "communication within Tester",
        "communication from Tester to Render, Displayer",
        "computation of Render",         
        "communication from Render to Displayer",
        "computation of Displayer",
        "whole system"  
};

enum NodeStatus                 
{
        MASTER, RENDER, TESTER, DISPLAYER
};



const static char* NodeName[4]=
{
        "Master Node",
        "Render Node",
        "Tester Node",
        "Displayer Node"
};

float total[4][MaxNodeNumber][ProfileItemCount]={0};
char nodeName[4][MaxNodeNumber][10];
int nodeCount[4]={0};


void display()
{
	int i, j, k;
#ifndef	DISPLAY_EXCEL
	for (i=0; i<4; i++)
	{
		printf("\n[%s][%d]\n", NodeName[i], nodeCount[i]);
		total[i][MaxNodeNumber-1][ALL]/=nodeCount[i];
		for (j=0; j<ProfileItemCount-1; j++)
		{
			if (total[i][MaxNodeNumber-1][j]!=0)
			{
				total[i][MaxNodeNumber-1][j]/=nodeCount[i];
				printf("[%s]=[%f][%f\%]\n", ProfileItemName[j], total[i][MaxNodeNumber-1][j],
				total[i][MaxNodeNumber-1][j]/total[i][MaxNodeNumber-1][ALL]*100.0);
			}
		}
		printf("[%s]=[%f]\n", ProfileItemName[ALL], total[i][MaxNodeNumber-1][ALL]);

	}
	printf("\n**************detail is following************************\n");
	for (i=0; i<4; i++)
	{
		printf("\n[%s]details of total[%d]\n", NodeName[i], nodeCount[i]);
		for (j=0; j<nodeCount[i]; j++)
		{
			printf("\n[%s#%d][%s]:\n",NodeName[i], j, nodeName[i][j]);
			for (k=0; k<ProfileItemCount-1; k++)
			{
				if (total[i][j][k]!=0)
				{
					printf("[%s]=[%.0f][%.2f\%]\n", ProfileItemName[k], total[i][j][k], 
						total[i][j][k]/total[i][j][ALL]*100.0);
				}
			}
			printf("[%s]=[%.0f]\n", ProfileItemName[ALL], total[i][j][ALL]);
		}
	}
#else
	for (i=0; i<4; i++)
	{
		total[i][MaxNodeNumber-1][ALL]/=nodeCount[i];
		for (j=0; j<nodeCount[i]; j++)
		{
			for (k=0; k<ProfileItemCount; k++)
			{
				printf("%.0f  ", total[i][j][k]);
			}
			printf("\n");
		}
	}
	printf("fps=%.2f\n", 500*1000000.0/total[0][MaxNodeNumber-1][ALL]);
#endif
}
		


void retrieveTerm(char* buf, char* term, char* value)
{
	char* startPos, *endPos;

	startPos=strchr(buf, '[');
	endPos=strchr(buf, ']');
	if (startPos==NULL || endPos==NULL)
	{
		perror("retrieve term error:");
		exit(2);
	}
	
	strncpy(term, startPos+1, endPos-startPos-1);
	term[endPos-startPos-1]='\0';
	//printf("term=%s\n", term);

	startPos=strchr(endPos, '[');
	//printf("startPos=%s\n", startPos);
	endPos=strchr(startPos, ']');
	//printf("endPos=%s\n", endPos);

	if (startPos==NULL||endPos==NULL)
	{
		perror("retrieve term error:");
		exit(2);
	}

	strncpy(value, startPos+1, endPos-startPos-1);
	value[endPos-startPos-1]='\0';
}

NodeStatus whichNode(char* buf)
{
	char term[256];
	char value[256];
	retrieveTerm(buf, term, value);
	//printf("whichNode:buf=%s,term=%s,value=%s\n", buf, term, value);
	for (int i=0; i<4; i++)
	{
		if (strcmp(term, NodeName[i])==0)
		{
			strcpy(nodeName[i][nodeCount[i]], value);
			return (NodeStatus)(i);
		}
	}
	printf("error! node status unknown\n");
	exit(1);
}

ProfileItem whichItem(char* term)
{
	int i;
	//printf("term=%s\n", term);
	for (i=0; i<ProfileItemCount; i++)
	{
		if (strcmp(term, ProfileItemName[i])==0)
		{
			break;
		}
	}
	if (i==ProfileItemCount)
	{
		perror("cannot find item:");
		exit(3);
	}
	return (ProfileItem)(i);
}
	
				
void processData(NodeStatus nodeStatus, char* buf)
{
	char term[256];
	char value[256];
	ProfileItem item;
	int i;
	retrieveTerm(buf, term, value);
	item=whichItem(term);
	sscanf(value, "%d", &i);
	total[nodeStatus][nodeCount[nodeStatus]][item]=i;
	total[nodeStatus][MaxNodeNumber-1][item]+=i;
}
		 	
	
void processFile(char* fileName)
{
	FILE* file;
	int pos, i;
	char buf[256];
	NodeStatus nodeStatus;
	if ((file=fopen(fileName, "r"))==NULL)
	{
		printf("cannot open file %s\n", fileName);
		exit(1);
	}
	//determine node status
	if (fgets(buf, 256, file)==NULL)
	{
		perror("read error:");
		
	}
	//printf("get first line %s\n", buf);
	
	nodeStatus=whichNode(buf);
	while (fgets(buf, 256, file))
	{
		processData(nodeStatus, buf);
	}	
	nodeCount[nodeStatus]++;

}			


void findAllFile(char* dirName)
{
        struct stat buf;
        char curFileName[256];
        int nameLen;
        int startPos;

        char* src, *tgt;
        DIR* dirPtr;
        struct dirent* direntPtr;
        //printf("begin searching %s...\n", dirName);

        if ((dirPtr=opendir(dirName))==NULL)
        {
                //perror("read dir");
                //exit(1234);
                //printf("unable open %s\n", dirName);
                return;
        }
        nameLen=strlen(dirName);
        strcpy(curFileName, dirName);
        curFileName[nameLen]='/';

        while ((direntPtr=readdir(dirPtr))!=NULL)
        {
		startPos=strlen(direntPtr->d_name)-4;
                if (direntPtr->d_name[0]=='n'&&strcmp(direntPtr->d_name+startPos, ".txt")==0)
                {
                        strcpy(curFileName+nameLen+1, direntPtr->d_name);
			//printf("find file %s\n", curFileName);
			processFile(curFileName);		
                }
        }
        closedir(dirPtr);
}

int main(int argc, char** argv)
{
	if (argc!=2)
	{
		printf("usage: sumTool.exe dirName\n");
		exit(0);
	}
	findAllFile(argv[1]);
	display();
	return 0;
}
