#include "viewCollector.h"


bool ViewCollector::fullRenderMode=false;
bool ViewCollector::hierRenderMode=true;
int ViewCollector::vertexCounter=0;
int ViewCollector::faceCounter=0;
Octree* ViewCollector::root=NULL;

void ViewCollector::startCollecting(VVector origin)
{
	frustum.CalculateFrustum();
	vertexCounter=faceCounter=0;
	octreeNodeVector.clear();
	if (hierRenderMode)
	{	
		collectHierLeafNode(0);
	}
	else
	{
		collectLeafNode(origin);
	}
}

void ViewCollector::addCandidate(int index)
{
	vertexCounter+=root[index].vNumber+root[index].vRepeatNumber;
	faceCounter+=root[index].fNumber+root[index].fRepeatNumber;
	//to do:
	//put into result set
	octreeNodeVector.push_back(index);
}

void ViewCollector::doCollectLeafNode(int index)
{
	Octree* me=root+index;
	int i;
	pair<OctreeNodeSet::iterator, bool> result;
	//we insert the searched path in the first place

	//no matter it is leaf node or not
	result=searchSet.insert(index);
	if (!result.second)
	{
		return;
	}

	if (!frustum.canCullCube(root[index].center[0], root[index].center[1], 
		root[index].center[2], root[index].width/2.0))
	{
		for (i=0; i<root[index].neighbourCount; i++)
		{			
			searchQueue.push_back(root[index].neighbours[i]);							
		}

		if (root[index].sonCount>0)
		{
			for (i=0; i<8; i++)
			{		
				if (root[index].sons[i]!=-1)
				{
					doCollectLeafNode(root[index].sons[i]);
				}
			}
		}
		else
		{
			//now they are leaves and we test and add
			//and they must already be added into searchSet!!!!!!
			//and  they must have already been added into nodeSet by testAndAdd
			//now we handle its neighbour immediately
			testAndAdd(index);		
		}
	}
}


//a typical BFS because we want to search from near to far within capacity
void ViewCollector::collectLeafNode(VVector origin)
{
	int i;
	vector<int> result;	

	//nodeQueue[nextIndex].clear();
	searchSet.clear();
	searchQueue.clear();
	

	if(!getStartingNodes(0, origin, searchQueue))
	{
		//we just add all and this will be a problem when viewer is outside octree boundbox
		for (i=0; i<root->infoStruct.nodeCount; i++)
		{
			if (root[i].sonCount==0)
			{
				//none of them has been tested
				searchQueue.push_back(i);				
			}
		}
	}


	//printf("searchQueue.size=%d\n", searchQueue.size());
	//let node decide whether to put itself or not
	while (!searchQueue.empty())
	{
		if (checkCapacity())
		{
			//don't put
			break;
		}		
		int index=searchQueue.front();
		searchQueue.pop_front();
		
		//so this queue is used as a searching clue, if the element itself
		//is inside frustum, we add it into our nodeSet[nextInt].
		//and also test its neighbours, provided they are not searched
		//if not, we don't even care about its neighbours
		//so, whatever is inside searchqueue are all our candidates for testing.
		doCollectLeafNode(index);//add more neighbours		
	}

}


//this is a bit tricky, because it is possible that there is no leaf nodes contain
//the point and the internal node contains the point, in this case we still need
//those leaf nodes which belongs to lowest internal node
bool ViewCollector::getStartingNodes(int index, VVector origin, OctreeNodeQueue& result)
{
	Octree* ptr;
	int i;
	ptr=root+index;		

	for (i=0; i<3; i++)
	{
		if (origin[i]<ptr->center[i]-ptr->width/2.0-Epsilon 
			||origin[i]>ptr->center[i]+ptr->width/2.0+Epsilon)
		{
			return false;//no condition
		}			
	}

	if (!frustum.canCullCube(ptr->center[0], ptr->center[1], ptr->center[2], ptr->width/2.0))
	{
		result.push_back(index);		
	}
	else
	{		
		//this is tricky, it means the point is contained in the cube
		//but our cube is not contained inside frustum or intersects with.
		//the only explanation is that the cube contains frustum, so go down
		
		for (i=0; i<8; i++)
		{
			if (ptr->sons[i]!=-1)
			{
				getStartingNodes(ptr->sons[i], origin, result);				
			}
		}	
	}
	//we must report adding successful
	return result.size()!=0;//no condition
	
}

bool ViewCollector::checkCapacity()
{
	return vertexCounter>OctreeLoader::capacityFactor*MaxVertexCapacity||
		faceCounter>OctreeLoader::capacityFactor*MaxFaceCapacity||
		octreeNodeVector.size()>MaxPossibleNodeCollectable;
}

void ViewCollector::collectHierLeafNode(int index)
{
	if (checkCapacity())
	{
		return;
	}
	if (!frustum.canCullCube(root[index].center, root[index].width/2.0))
	{
		if (root[index].sonCount==0)
		{
			addCandidate(index);		
		}
		else
		{
			for (int i=0; i<8; i++)
			{
				if (root[index].sons[i]!=-1)
				{
					collectHierLeafNode(root[index].sons[i]);
				}
			}
		}
	}
}


bool ViewCollector::testAndAdd(int nodeIndex)
{
	Octree*ptr=root+nodeIndex;

	if (!frustum.canCullCube(ptr->center[0], ptr->center[1], ptr->center[2], ptr->width/2.0))
	{  	
		//we only counts the leaf node data
		addCandidate(nodeIndex);		
		return true;
	}   
	else
	{
		return false;
	}
}
